Function bottom,f,lambda1,ITER=iter,EPS=eps,POLY=pol,X=xarg $
               ,MIN=mn1,MAX=mx1,LAM2=lambda2,WEIGHT=weight
; bottom tries to fit a smooth curve to the lower envelope
; of 1D data array f. Filter size "filter"
; together with the total number of iterations determine
; the smoothness and the quality of the fit. The total
; number of iterations can be controlled by limiting the
; maximum number of iterations (iter) and/or by setting
; the convergence criterion for the fit (eps)
; 04-Nov-2000 N.Piskunov wrote.
; 09-Nov-2011 NP added weights and 2nd derivative constraint as LAM2
;
common reducelib, reducelib_name
  if(n_params() lt 2) then begin
    print,'syntax: bottom,f,{filter/order}[,X=xarg,[ITER=iter[,EPS=eps $'
    print,'             [,MIN=mn[,[MAX=mx[,/POLY[,LAM2=lam2[,WEIGHT=wgt]]]]]]]]'
    print,'where f      is the function to fit,'
    print,'      filter is the smoothing parameter for the optimal filter.'
    print,'             If POLY is set, it is interpreted as the order'
    print,'             of the smoothing polynomial,'
    print,'      xarg   optional X-coordinates of data points,'
    print,'      iter   is the maximum number of iterations [def: 40]'
    print,'      eps    is convergence level [def: 0.001]'
    print,'      mn     minimum function values to be considered [def: min(f)]'
    print,'      mx     maximum function values to be considered [def: max(f)]'
    print,'      lam2   constraint on 2nd derivative'
    print,'      wgt    vector of weights.'
    return,0
  endif
  if(not keyword_set(iter)) then iter=40
  if(not keyword_set(eps)) then eps=0.001
  if(not keyword_set(mn1)) then mn=min(f) else mn=mn1
  if(not keyword_set(mx1)) then mx=max(f) else mx=mx1

  if(n_elements(lambda1) lt 1) then begin
    print,'bottom: smoothing parameter should be defined.'
    return,f
  endif else if(lambda1 lt 0) then begin
    print,'bottom: smoothing parameter should not be negative.'
    return,f
  endif else filter=double(lambda1)

  if(n_elements(lambda2) lt 1) then begin
    lam2=0
  endif else if(lambda2 lt 0) then begin
    lam2=0
  endif else lam2=double(lambda2)

  if(keyword_set(pol)) then begin
    j=where(f ge mn and f le mx, n)
    if(n le round(filter)) then return,f
    if(n_elements(f) eq n_elements(xarg)) then begin
      xx=2.d0*(xarg-(max(xarg)+min(xarg))*0.5)/(max(xarg)-min(xarg))
    endif else begin
      xx=(2*dindgen(n_elements(f))/(n_elements(f)-1)-1)
    endelse
    fmin=min(f(j))-1
    fmax=max(f(j))+1
    ff=(f(j)-fmin)/(fmax-fmin)
    ff_old=ff
  endif else begin
    fff=middle(f,filter,X=xarg,ITER=iter,EPS=eps,WEIGHT=weight,LAM2=lam2)
    fmin=min(f)-1
    fmax=max(f)+1
    fff=(fff-fmin)/(fmax-fmin)
    ff=(f-fmin)/(fmax-fmin)/fff
    ff_old=ff
    n=n_elements(f)
    if(n_elements(xarg) eq n) then begin
      xx=double(xarg)
    endif
  endelse
  if(keyword_set(weight)) then wgt=weight else wgt=1

  if(not keyword_set(reducelib_name)) then begin
    help,calls=a
    a=strmid(a(0),strpos(a(0),'<')+1,strlen(a(0))-strpos(a(0),'<')-1)
;    if(strpos(a,'/nfs') eq 0) then a=strmid(a,4)
    if(strpos(a,'/export') eq 0) then a=strmid(a,7)
    if(!VERSION.OS eq 'Win32') then delimiter='\' else delimiter='/'
    prefix=strmid(a,0,strpos(a,delimiter,/REVERSE_SEARCH))   ;slit_func directory
    if(strpos(a,delimiter) lt 0) then cd,CURRENT=prefix
    reducelib_name =prefix+delimiter+'reduce.so.'   $
                                    +!version.os+'.' $
                                    +!version.arch+'.' $
                                    +strtrim(!version.MEMORY_BITS,2)
  endif

  Opt=strarr(3)
  nopt=0
  if(n_elements(xarg)  eq n) then begin
    Opt[nopt]='Xarg'
    nopt=nopt+1
  endif
  if(n_elements(wgt)   eq n) then begin
    Opt[nopt]='Weights'
    nopt=nopt+1
  endif
  if(keyword_set(lam2)) then begin
    Opt[nopt]='Lam2'
    nopt=nopt+1
  endif

  i=0
next:i=i+1
  if(keyword_set(pol)) then begin
    order=round(filter)
    if(order gt 0) then begin ; This is a bug in RSI poly routine
      t=median(poly(xx,poly_fit(xx,ff,order,/DOUBLE)),3)
      dev=sqrt(poly(xx,poly_fit(xx,((t-ff)>0.)^2,order,/DOUBLE))>0.)
    endif else begin
      t=replicate(poly_fit(xx,ff,order,/DOUBLE),n_elements(f))
      dev=sqrt(replicate(poly_fit(xx,((t-ff)>0.)^2,order,/DOUBLE),n_elements(f))>0.)
    endelse
  endif else begin
    if(nopt eq 3) then begin
      t=dblarr(n)
      iret=call_external(reducelib_name, 'opt_filter_1d', ff, t, n, filter, Opt, xx, wgt, lam2)
      if(iret ne 0) then begin
        print, iret
        stop
      endif
      t=median(t, 3)
      dev=dblarr(n)
      iret=call_external(reducelib_name, 'opt_filter_1d', wgt*(t-ff)>0.d0, dev, n, filter, Opt, xx, wgt, lam2)
      dev=sqrt(dev>0.d0)
    endif else if(nopt eq 2) then begin
      if(Opt[0] eq 'Xarg' and Opt[1] eq 'Weights') then begin
        t=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', ff, t, n, filter, Opt, xx, wgt)
        if(iret ne 0) then begin
          print, iret
          stop
        endif
        t=median(t, 3)
        dev=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', wgt*(t-ff)>0.d0, dev, n, filter, Opt, xx, wgt)
        dev=sqrt(dev>0.d0)
      endif else if(Opt[0] eq 'Xarg' and Opt[1] eq 'Lam2') then begin
        t=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', ff, t, n, filter, Opt, xx, lam2)
        if(iret ne 0) then begin
          print, iret
          stop
        endif
        t=median(t, 3)
        dev=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', wgt*(t-ff)>0.d0, dev, n, filter, Opt, xx, lam2)
        dev=sqrt(dev>0.d0)
      endif else if(Opt[0] eq 'Weights' and Opt[1] eq 'Lam2') then begin
        t=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', ff, t, n, filter, Opt, wgt, lam2)
        if(iret ne 0) then begin
          print, iret
          stop
        endif
        t=median(t, 3)
        dev=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', wgt*(t-ff)>0.d0, dev, n, filter, Opt, wgt, lam2)
        dev=sqrt(dev>0.d0)
      endif
    endif else if(nopt eq 1) then begin
      if(Opt[0] eq 'Xarg') then begin
        t=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', ff, t, n, filter, Opt, xx)
        if(iret ne 0) then begin
          print, iret
          stop
        endif
        t=median(t, 3)
        dev=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', wgt*(t-ff)>0.d0, dev, n, filter, Opt, xx)
        dev=sqrt(dev>0.d0)
      endif else if(Opt[0] eq 'Weights') then begin
        t=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', ff, t, n, filter, Opt, wgt)
        if(iret ne 0) then begin
          print, iret
          stop
        endif
        t=median(t, 3)
        dev=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', wgt*(t-ff)>0.d0, dev, n, filter, Opt, wgt)
        dev=sqrt(dev>0.d0)
      endif else if(Opt[0] eq 'Lam2') then begin
        t=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', ff, t, n, filter, Opt, lam2)
        if(iret ne 0) then begin
          print, iret
          stop
        endif
        t=median(t, 3)
        dev=dblarr(n)
        iret=call_external(reducelib_name, 'opt_filter_1d', wgt*(t-ff)>0.d0, dev, n, filter, Opt, lam2)
        dev=sqrt(dev>0.d0)
      endif
    endif else if(nopt eq 0) then begin
      t=dblarr(n)
      iret=call_external(reducelib_name, 'opt_filter_1d', ff, t, n, filter, Opt)
      if(iret ne 0) then begin
        print, iret
        stop
      endif
      t=median(t, 3)
      dev=dblarr(n)
      iret=call_external(reducelib_name, 'opt_filter_1d', wgt*(t-ff)>0.d0, dev, n, filter, Opt)
      dev=sqrt(dev>0.d0)
    endif
  endelse
  ff=((t-dev)>ff)<t
  dev2=max(wgt*abs(ff-ff_old))
  ff_old=ff
  if(dev2 gt eps and i lt iter) then goto,next

  if(keyword_set(pol)) then begin
    x=(2*dindgen(n_elements(f))/(n_elements(f)-1)-1)
    if(order gt 0) then begin ; This is a bug in RSI poly routine
      t=median(poly(xx,poly_fit(xx,ff,order,/DOUBLE)),3)
    endif else begin
      t=replicate(poly_fit(xx,ff,order,/DOUBLE),n_elements(f))
    endelse
    return,t*(fmax-fmin)+fmin
  endif else $
    return,t*fff*(fmax-fmin)+fmin
end
