pro get_shear,ref_image,delta_x,shear_x,save_file=save_file,POLY=poly,PLOT=plot $
             ,MASK=mask,inst_mode=inst_mode,exten=exten,POLARIZ=polariz

  if(n_params() lt 1) then begin
    print,'Syntax: get_shear,ref_image or ref_file,[save_file=save_file[' $
         +',inst_mode=inst_mode[,POLY=poly[,POLARIZ=pol[,/PLOT]]]]]'
    print,'   thar_filename - name of the ThAr FITS file'
    print,'   save-file     - REDUCE save file containing order locations'
    print,'   inst_mode     - REDUCE instrument setup'
    print,'   poly          - polynomial order to fit the tilt as function of column number'
    print,'   pol           - 2/1, if each order is reproduced twice as in polarimetry'
    print,'                   one can select just one lower (2) or upper (1) image'
    print,'   /PLOT         - a flag plotting intermediate fits'
    return
  endif

; Read reference image
  if(not keyword_set(inst_mode)) then begin
    print,'GET_SHEAR: inst_mode must be defined'
    return
  endif
  if not keyword_set(poly) then poly=2
  s=size(ref_image)
  if s[0] eq 0 and s[1] eq 7 then begin ; We have a string, probably filename
    fits_read,ref_image,ThAr,hdr,exten=exten,/PDU
    thar_file=median(ThAr,5)
    hdr=modeinfo(hdr,inst_mode)
    ThAr=clipnflip(ThAr,hdr)
    thar_filename=ref_image
  endif else if s[0] eq 2 then begin   ; We have a 2D image
    ThAr=median(ref_image,5)
    thar_filename='thar.fits'
  endif

  if n_elements(ThAr) eq n_elements(mask) then ThAr=ThAr*mask
  ncol=n_elements(ThAr[*,0])
  nrow=n_elements(ThAr[0,*])
;  bckg=median(ThAr[*,6:43],dimen=2)
;  bckg=smooth(bckg,3)
;  ThAr=ThAr-(bckg#replicate(1,nrow))
;  ThAr=median(ThAr, 3)

; Restore order location
  if(keyword_set(save_file)) then restore,save_file $
  else begin
    print,'GET_SHEAR: Save file with order fit and column ranges must be specified'
    return
  endelse

  if keyword_set(polariz) then begin
    nord=n_elements(orders[0,*])
    i=dindgen(nord/2)*2+(polariz mod 2)
    orders=orders[*,i]
    col_range=col_range[*,i]
    def_xwd  =def_xwd[*,i]
  endif

  nord=n_elements(orders[0,*])
;  i=indgen(nord/2)*2
;  orders=orders[*,i]
;  col_range=col_range[*,i]
;  nord=nord/2

; Make symmetric extraction window and reduce it a bit
  xwd=min(def_xwd)-1
;  xwd=5

; Slit shearing will be fit with parabola as function of pixel number
  shear_x=dblarr(poly+1,nord)

  if(keyword_set(plot)) then colors
  for iord=0,nord-1 do begin ; Go through sp. orders
;if iord ge 60 then plot=1

; Local strongest sp. lines
    sp=dblarr(ncol)
    ycen=poly(dindgen(ncol),orders[*,iord])
    for i=col_range[0,iord],col_range[1,iord] do begin
      sp[i]=total(ThAr[i,ycen[i]-xwd:ycen[i]+xwd])
    endfor
    mean_sp=bottom(sp[col_range[0,iord]:col_range[1,iord]],1.d3)
    i=where(sp[col_range[0,iord]:col_range[1,iord]] lt mean_sp, ni)
    dev=sqrt(mean(((sp[col_range[0,iord]:col_range[1,iord]]-mean_sp)^2)[i]))
;    mean_sp=bottom(sp[col_range[0,iord]:col_range[1,iord]],1.d3)
    if(keyword_set(plot)) then begin
      plot,sp,xs=1,ys=3
      oplot,indgen(col_range[1,iord]-col_range[0,iord])+col_range[0,iord],mean_sp,line=3
    endif
    locmax=where(sp[1:ncol-2] gt sp[0:ncol-3] and $
                 sp[1:ncol-2] gt sp[2:ncol-1], nmax)+1

    igood=where(sp[locmax] gt mean_sp[locmax-col_range[0,iord]]+dev*3, ngood)
    if ngood gt 2 then begin
      locmax=locmax[igood]
      nmax=ngood
    endif

    if nmax lt 2 then continue

; Mark lines which are too close to each other to confuse Gauss fit
    ibad=where(locmax[1:nmax-1]-locmax[0:nmax-2] lt 20, nbad)
    while(nbad gt 0) do begin

; For bad pairs keep the strongest of the two lines
      for i=0,nbad-1 do $
        if sp[locmax[ibad[i]+1]] lt sp[locmax[ibad[i]]] then ibad[i]=ibad[i]+1
      locmax[ibad]=-1
      locmax=locmax[where(locmax ge col_range[0,iord]+10 and $
                          locmax lt col_range[1,iord]-10, nmax)]
      ibad=where(locmax[1:nmax-1]-locmax[0:nmax-2] lt 20, nbad)
    endwhile

    if(keyword_set(plot)) then print,'Found '+strtrim(nmax,2)+' lines in order=',iord+1

; Keep not more than 201 strongest lines per order
    i=reverse(sort(sp[locmax]))
    i=i[0:40<(nmax-1)]
    nmax=n_elements(i)
    locmax=locmax[i]
    i=sort(locmax)
    locmax=locmax[i]
    if(keyword_set(plot)) then begin
      oplot,locmax,sp[locmax],psym=2
      empty
      sss=get_kbrd(1)
    endif

    xx=dindgen(21)-10.d0  ; Gaussian fit will be based on 25 pixels
    xcen=dblarr(2*xwd+1)  ; Line centers will be stored along the slit
    shear=dblarr(nmax)    ; Shearing will be determined for each line separately

    ibad=intarr(nmax)
    for iline=0,nmax-1 do begin ; Loop through the lines
      x=locmax[iline]+xx
      for irow=-xwd,xwd do begin
        s=ThAr[x,irow+ycen[locmax[iline]]] ; Extract short horizontal strip
        yy=disp_gaussfit(xx,s,a,xx)        ; Do Gauss fit
        xcen[irow+xwd]=locmax[iline]+a[1]  ; Store line center
      endfor
      if(keyword_set(plot)) then plot,xcen,ys=3,psym=2,tit='Order ' $
                                      +strtrim(iord+1,2)+' line ' $
                                      +strtrim(iline+1,2)+' out of ' $
                                      +strtrim(nmax,2) $
                                      ,xtit='Slit position' $
                                      ,ytit='ThAr line position',xs=3
      a=poly_fit(dindgen(2*xwd+1)-xwd,xcen,1,/double) ; Linear fit to slit image
      j=sort(abs(poly(dindgen(2*xwd+1)-xwd,a)-xcen))  ; Outlyers
      a=poly_fit(double(j[0:2*xwd*0.8])-xwd $         ; Fit again using best 80%
                  ,xcen[j[0:2*xwd*0.8]],1,/DOUBLE)
      j=sort(abs(poly(dindgen(2*xwd+1)-xwd,a)-xcen))  ; Outlyers again
      if(keyword_set(plot)) then oplot,j[0:2*xwd*0.8] $
                            ,xcen[j[0:2*xwd*0.8]],psym=2,col=3
      a=poly_fit(double(j[0:2*xwd*0.8])-xwd $         ; Final fit using best 80%
                  ,xcen[j[0:2*xwd*0.8]],1,/DOUBLE)
      resid=xcen-poly(dindgen(2*xwd+1)-xwd,a)
      delta=total(resid^2)/(2*xwd-1)/stddev(dindgen(2*xwd+1)-xwd)
      if delta gt 0.3*stddev(xcen) then ibad[iline]=1
      print,iline,delta,stddev(xcen),dev

      if(keyword_set(plot)) then oplot,poly(dindgen(2*xwd+1)-xwd,a)
      if(keyword_set(plot)) then empty
      shear[iline]=a[1]                               ; Store line shear

      if ibad[iline] eq 0 then begin
        oplot,[-xwd,xwd]*a[1]+locmax[iline],[-xwd,xwd]+ycen[locmax[iline]],col=c24(3),thick=2
;        stop
      endif
    endfor
    igood=where(ibad eq 0, nmax)
    locmax=locmax[igood]
    shear=shear[igood]
    if nmax le 1 then stop

    if nmax eq 2 then begin
      shear_x[0:1,iord]=poly_fit(locmax[0:nmax-1], shear, 1, /DOUBLE)
      continue
    endif
    if(keyword_set(plot)) then plot,locmax,shear,ys=3,xs=1,psym=2 $
                                   ,tit='Slit tilt for order ' $
                                   +strtrim(iord+1,2),xtit='Pixel #' $
                                   ,ytit='Slit tilt'
    a=poly_fit(locmax,shear,1,/double)
    j=sort(abs(poly(locmax,a)-shear))
    a=poly_fit(locmax[j[0:nmax*0.7]],shear[j[0:nmax*0.7]],poly,/DOUBLE)
    j=sort(abs(poly(locmax,a)-shear))
    if(keyword_set(plot)) then oplot,locmax[j[0:nmax*0.7]] $
                                   ,shear[j[0:nmax*0.7]],psym=2,col=3
    a=poly_fit(locmax[j[0:nmax*0.8]],shear[j[0:nmax*0.8]],poly,/DOUBLE)
    if(keyword_set(plot)) then oplot,locmax,poly(locmax,a)
    if(keyword_set(plot)) then empty
    shear_x[*,iord]=a
    if(keyword_set(plot)) then ss=get_kbrd(1)
  endfor

; Fix runners among sp. orders
  x=dindgen(nord)
  for ii=0,poly do begin
    y=shear_x[ii,*]
    yfit=poly(x,poly_fit(x,y,1,/DOUBLE))
    dev=sqrt(mean((y-yfit)^2))
    i=where(abs(y-yfit) lt 5*dev, ni)
    if(ni lt poly) then begin
      print,'GET_SHEAR: We are in trouble!'
      stop
    endif
    yfit=poly(x,poly_fit(x[i],y[i],1,/DOUBLE))
    shear_x[ii,*]=yfit
    if keyword_set(plot) then begin
      plot,x,y,psym=2,tit='Coefficient '+strtrim(ii,2)+' for polynomial fit to the slit tilt'
      oplot,x,yfit,psym=2,col=c24(2)
      sss=get_kbrd(1)
    endif
  endfor

  delta_x=intarr(nord)
  for i=0,nord-1 do delta_x[i]=ceil(max(abs(poly([0,ncol] $
                                   ,shear_x[*,i]))*(xwd+2.)))
;  iord=indgen(33)*2+1
;  nord=n_elements(iord)
;  delta_x=delta_x[iord]
;  shear_x=shear_x[*,iord]

  f=strmid(thar_filename,0,strpos(thar_filename,'fits')) $
           +inst_mode+'_slit_tilt.sav'
  save,delta_x,shear_x,file=f
  print,'Slit tilt functions and the PSF windows are stored in '+f
  
  return
end
