function nonlinear_saao_giraffe,image,header,AMPLIFIER=amplifier,GAIN=gain

; Readout speed affects non-linearity correction
  rdspeed = strtrim(sxpar(header, 'SPEEDMOD', count=count), 2)

; Check the number of amplifiers used
  ver = sxpar(header, 'E_HVERS', count=count)
  if(ver gt 1.001) then begin
    n_amp = sxpar(header, 'E_AMPL', count=count)
  endif else begin
    n_amp = 1
  endelse

; For image fragment user must specify the amplifier and the gain
; Only one set of values is allowed as we do not have geometrical
; reference. This type of call is made from inside sumfits.
  if(keyword_set(amplifier) and keyword_set(gain)) then begin
    return, image * gain
  endif

; Alternative mode: process the whole image at once
  if(n_amp gt 1) then begin
    image=float(image)
    for amp=1,n_amp do begin
      sf = suffix(amp, 1)
      xlo  = sxpar(header, 'E_XLO' +sf, count=count)
      ylo  = sxpar(header, 'E_YLO' +sf, count=count)
      xhi  = sxpar(header, 'E_XHI' +sf, count=count)
      yhi  = sxpar(header, 'E_YHI' +sf, count=count)
      gain = sxpar(header, 'E_GAIN'+sf, count=count)
      image[xlo:xhi,ylo:yhi] = image[xlo:xhi,ylo:yhi] * gain
    endfor
  endif

  return,image
end

pro modeinfo_saao_giraffe,mode,newhead,object,orient,xlo,xhi,ylo,yhi,gain,readn,backg,time $
                        ,obslat,obslon,obsalt,ra2000,de2000,jd,NONLINEAR=nonlinear
;===========================================================================================
;The CTIO Chiron spectrometer (2-amplifier mode)

  reorient = 2 ; Wavelength runs right to left and top to bottom
  if n_elements(orient) eq 0 then orient = reorient

  dsec_bias = hierarch(newhead, 'BIASSEC', count=count, errmes=errmes)
  if count eq 0 then message, errmes
  i1 = strpos(dsec_bias,'[') & i2 = strpos(dsec_bias,']')
  dsec_bias = strsplit(strmid(dsec_bias, i1+1, i2-i1-1),'[:,]',/EXTRACT)
  x0=fix(dsec_bias[0])
  x1=fix(dsec_bias[1])
  y0=fix(dsec_bias[2])
  y1=fix(dsec_bias[3])

  dsec_total = hierarch(newhead, 'TRIMSEC', count=count, errmes=errmes)
  if count eq 0 then message, errmes
  i1 = strpos(dsec_total,'[') & i2 = strpos(dsec_total,']')
  dsec_total = strsplit(strmid(dsec_total, i1+1, i2-i1-1),'[:,]',/EXTRACT)
  x0_bottom=fix(dsec_total[0])
  x1_bottom=fix(dsec_total[1])-2
  y0_bottom=1
  y1_bottom=511

  x0_top=fix(dsec_total[0])
  x1_top=fix(dsec_total[1])-2
  y0_top=512
  y1_top=1022

  xlo = [x0_bottom, x0_top]
  xhi = [x1_bottom, x1_top]
  ylo = [y0_bottom, y0_top]
  yhi = [y1_bottom, y1_top]

  if keyword_set(xr) then begin
    xlo = xr(0)
    xhi = xr(1)
  endif
  if keyword_set(yr) then begin
    ylo = yr(0)
    yhi = yr(1)
  endif

  if n_elements(gain) eq 0 then begin
;    gain_bottom = 1.d0/0.926
;    gain_top    = 1.d0/0.833
    gain_bottom = 0.926
    gain_top    = 0.833
    gain = [gain_bottom, gain_top]
  endif

  if n_elements(readn) eq 0 then begin
    readn_bottom = 6.1
    readn_top = 6.7
    readn = [readn_bottom, readn_top]
    if count eq 0 then message, errmes
  endif

  if n_elements(backg) eq 0 then begin
    drk = 0.
    sky = 0.
    backg = gain * (drk + sky)        ;convert ADU to electrons
  endif

  if n_elements(time) eq 0 then begin
    time = hierarch(newhead, 'EXPTIME', count=count, errmes=errmes)
    if count eq 0 then message, errmes
  endif

; For object frames, prepare for heliocentric correction
  imtype=hierarch(newhead,'OBJECT',count=count,errmes=errmes)
  imtype=strtrim(imtype, 2)
  if(imtype ne 'ARC' and imtype ne 'CAMERA' and imtype ne 'FLAT') then begin
    if n_elements(ra2000) eq 0 then begin
      ra = hierarch(newhead,'RA',count=count, errmes=errmes)
      ra = float(strsplit(ra,':',/EXTRACT))
      ra = ten(ra[0],ra[1],ra[2])
      ra2000 = ra/3600.
    endif
    if n_elements(de2000) eq 0 then begin
      dec = hierarch(newhead,'DEC',count=count, errmes=errmes)
      dec = float(strsplit(dec,':',/EXTRACT))
      dec = ten(dec[0],dec[1],dec[2])
      de2000 = dec/3600.
    endif
    if n_elements(jd) eq 0 then begin
      date_obs = hierarch(newhead, 'DATE-OBS', count=count, errmes=errmes)
      ut = hierarch(newhead, 'TIME-OBS', count=count, errmes=errmes)
      date_obs = strsplit(date_obs, '-', /EXTRACT)
      ut = float(strsplit(ut, ':', /EXTRACT))
      ut = ut[0] + ut[1] / 60.d0 + ut[2] / 3600.d0
      jdcnv,date_obs[0], date_obs[1], date_obs[2], ut, jd  ;convert UT to JD
      jd = jd - 2400000.d0 + time * 0.5d0
    endif
;observatory coordinates
    if n_elements(obslon) eq 0 then obslon = -18.4771d0
    if n_elements(obslat) eq 0 then obslat = -33.9345d0
    if n_elements(obsalt) eq 0 then obsalt = 1759.d0
  endif
  nonlinear='nonlinear_saao_giraffe'
  return
end
